unit Coder;

interface

type
  TCoderInpDataType = byte; //  !
  TCoderInpDataLen = cardinal;

const
  TCoderInpDataTypeSize = sizeof(TCoderInpDataType);
  TCoderInpDataLenSize = sizeof(TCoderInpDataLen);

type
  TCoderInpData = array [word] of TCoderInpDataType;
  PCoderInpData = ^TCoderInpData;

type
  TCoder = class
  private
    FBits: cardinal;
    Ftablesize: TCoderInpDataLen;
    FLimit: TCoderInpDataType; //   
    FHighLimit: TCoderInpDataType;  //   

    freq:     array [TCoderInpDataType] of TCoderInpDataLen;
    procedure QuickSortFreq;
  public
    valuePos: array [TCoderInpDataType] of TCoderInpDataType; //    
    value: array [TCoderInpDataType] of TCoderInpDataType;    //   
    code:  array [TCoderInpDataType] of cardinal;             //   
    len:   array [TCoderInpDataType] of byte;                 //  

    property LowIndex: TCoderInpDataType read FLimit; //   
    property HighIndex: TCoderInpDataType read FHighLimit; //   

    property Bits: cardinal read FBits;  //  -    
    property Tablesize: TCoderInpDataLen read FTablesize;

    procedure Init(A: PCoderInpData; L,H: TCoderInpDataLen); //  
    procedure Process(len_inc: cardinal);
    function  BestLen: cardinal; //         
  end;

implementation

// *****************************************************************************
{ TCoder }

procedure TCoder.QuickSortFreq;
  procedure QSort(iLo, iHi: integer);
  var
     Lo, Hi: integer;
     Mid, T: TCoderInpDataLen;
  begin
   Lo:= iLo; Hi:= iHi; Mid:= freq[(Lo + Hi) div 2];
   repeat
     while freq[Lo] < Mid do Inc(Lo);
     while freq[Hi] > Mid do Dec(Hi);
     if Lo <= Hi then begin
       T:= freq[Lo]; freq[Lo]:= freq[Hi]; freq[Hi]:= T;
       T:= value[Lo]; value[Lo]:= value[Hi]; value[Hi]:= T;
       Inc(Lo); Dec(Hi);
     end;
   until Lo > Hi;
   if Hi > iLo then QSort(iLo, Hi);
   if Lo < iHi then QSort(Lo, iHi);
  end;
begin
  QSort(Low(freq), High(freq));
end;

procedure TCoder.Init(A: PCoderInpData; L,H: TCoderInpDataLen);
var i, j: integer;
    ii: TCoderInpDataLen;
begin
  FHighLimit:= High(Code);
  FillChar(code,sizeof(code),$0); FillChar(len,sizeof(len),$0);
  FillChar(valuePos,sizeof(valuePos),$0); FillChar(freq,sizeof(freq),$0);
  if not Assigned(A) then exit;

  for i:= Low(value) to High(value) do value[i]:= i;
  for ii:= L to H do inc(freq[A^[ii]]);
  QuickSortFreq;
  Flimit:= High(freq);
  while (freq[Flimit-1]<>0) and (Flimit>Low(freq)) do dec(Flimit);

  FTablesize:= (HighIndex-LowIndex+1)*2; FBits:= 0;
  for i:= High(valuePos) downto Low(valuePos) do //   
    for j:= High(value) downto Low(value) do
      if value[j]=i then begin
       valuePos[i]:= j;
       break;
      end;
end;

//  code  len  
procedure TCoder.Process(len_inc: cardinal);
var
  i: integer;
  curlen,
  prefixlen : byte;
  subcode, prefix : cardinal;
begin
  //    ( )   
  FBits:= 0;
  if LowIndex<HighIndex then begin
    while cardinal(HighIndex-LowIndex)<cardinal(1 shl (len_inc-1)) do dec(len_inc);

    curlen:= len_inc; subcode:= 0; prefixlen:= 0; prefix:= 0;
    for i:= HighIndex downto LowIndex do begin
     if ((subcode+1) shr (curlen - prefixlen) >0) and {    }
        (i<>LowIndex)  then  begin
       prefixlen:= curlen;
       prefix:= (1 shl prefixlen) - 1;
       curlen:= prefixlen+curlen*3 div 2;
       while cardinal(i-LowIndex+1)<cardinal(1 shl (curlen - prefixlen-1)) do
         dec(curlen);
       subcode:= 0;
     end;
     code[i]:= prefix shl (curlen - prefixlen) + subcode;
     inc(subcode);
     len[i]:= curlen;
     FBits:= FBits + freq[i]*curlen;
    end;
  end
  else begin //      1 )))))
    code[High(code)]:= 0;
    len[High(code)]:= 1;
  end;
end;

function TCoder.BestLen: cardinal; {returns len_inc}
var
  leninc, minsize: cardinal;
begin
  Result:= 1;
  minsize:= MaxInt;
  for leninc:= 1 to 10 do begin
     Process(leninc);
     if Bits<minsize then begin
       minsize:= bits;
       Result:= leninc;
     end;
  end;
  Process(Result);
end;

end.
