//************************************
//   Small but rather fine
//   compressor/decompressor
//   for 16bit data samples
//
//   version 1.01
//
//   E. Averchenkov (c) 1998-2000
//   e_g_o_r@mail.ru
//************************************
unit compressor;

interface

type
  InpDataType = smallint;  // type of input data
  InpDataLen = cardinal;   // type of input data index

type
  ByteArray = array [word] of byte;
  ByteArrayPtr = ^ByteArray;

  SmallIntArray = array [word] of InpDataType;
  SmallIntArrayPtr = ^SmallIntArray;

  function IsArrayCompressed(var Source: ByteArrayPtr; count: InpDataLen): boolean;

  function CompressArray(var Source: SmallIntArrayPtr; count: InpDataLen; var Dest: ByteArrayPtr; Compression: boolean): cardinal;
  function UnCompressArray(var Source: ByteArrayPtr; var Dest: SmallIntArrayPtr): InpDataLen;

implementation

uses Coder;

const
  InpDataTypeSize = sizeof(InpDataType);
  InpDataLenSize = sizeof(InpDataLen);

{     (offsets)}
{   -  16 }
{        2 }
{        }
{          }
{       }
const
  Signature = 'Compressed';

const
  Stored = 0; { } Compressed=1; {  }

  TypeOfData = Length(Signature)+1;{8 }
  DataWidth = TypeOfData+1;{    }

  Delta = DataWidth+1; 	{  0 -     }
  FirstWord = Delta+1;  {  Delta<>0 -   }

  DataBeg   = FirstWord+InpDataTypeSize; {   }
  DataCount = DataBeg+InpDataLenSize;   {     2 }
  Table1Beg = DataCount+InpDataLenSize; {      }
  Table2Beg = Table1Beg+InpDataLenSize; {      }
  BegOfDescrTable = Table2Beg+InpDataLenSize;
  {    }
  //   -     
  //    
  //    
    NumOfLenghts = 0; {8  -      }
    DescrTable = NumOfLenghts+1;
    {  }

function IsArrayCompressed(var Source: ByteArrayPtr; count: InpDataLen): boolean;
var i: integer;
begin
  Result:= false;
  if not Assigned(Source) or (count<BegOfDescrTable) then exit;
  Result:= true;
  for i:= 1 to Length(Signature) do
   Result:= Result and (Source^[Low(Source^)+i-1] = byte(Signature[i]));
end;


// *****************************************************************************
{ Compressor }

function CompressArray(var Source: SmallIntArrayPtr; count: InpDataLen; var Dest: ByteArrayPtr; Compression: boolean): cardinal;
var
  pointb, point1, point : cardinal;
  bitpointer: shortint;

  procedure WriteBits(Bits: cardinal; NumberOfBits: byte);
  var
    Nbit: byte;
  begin
   for Nbit:= NumberOfBits-1 downto 0 do begin
     if (cardinal(1 shl Nbit) and (Bits))<>0 then
       Dest^[pointb]:= Dest^[pointb] or (1 shl bitpointer);
     dec(bitpointer);
     if (bitpointer<0) then begin
       inc(pointb);
       bitpointer:= 7;
     end;
   end;
  end;

  procedure WriteByte(b: byte);
  begin
    Dest^[point]:= b;  inc(point);
  end;

  procedure WriteWord(addr: cardinal; wrd: word);
  begin
    point:= addr;
    WriteByte(Lo(wrd));  WriteByte(Hi(wrd));
  end;

  procedure WriteCardinal(addr: cardinal; x: cardinal);
  begin
    WriteWord(addr, word(x));  WriteWord(addr+sizeof(word), word(x div $10000));
  end;

  function ReadWord(pointr: cardinal): word;
  begin
    Result:= Dest^[pointr] + Dest^[pointr+1]*256;
  end;

  function ReadCardinal(pointr: cardinal): cardinal;
  begin
    Result:= Dest^[pointr] + Dest^[pointr+1]*$100 + Dest^[pointr+2]*$10000 + Dest^[pointr+3]*$1000000;
  end;

var
  i,
  deltas1size, deltas2size,
  lencount,  oldlen: cardinal;

  pos: TCoderInpDataType;

  tempValue: InpDataType;

  hf1, hf2,
  hfd1, hfd2,
  hfdd1, hfdd2 : TCoder;

  UseDeltas: byte;

var
   p1,p2 : PCoderInpData;
begin
  Result:= 0;
  if not Assigned(Source) then exit;
  UseDeltas:= 0;

  hf1:= TCoder.Create;
  hf2:= TCoder.Create;
  hfd1:= TCoder.Create;
  hfd2:= TCoder.Create;
  hfdd1:= TCoder.Create;
  hfdd2:= TCoder.Create;
  try
    //  
    GetMem(p1, count*sizeOf(Low(p1^)) ); FillChar(p1^,count*sizeOf(Low(p1^)),$0);
    GetMem(p2, count*sizeOf(Low(p2^)) ); FillChar(p2^,count*sizeOf(Low(p1^)),$0);
    try
      for i:= Low(Source^) to Low(Source^)+count-2 do begin
        p1^[i]:= Hi(word(Source^[i+1]));
        p2^[i]:= Lo(word(Source^[i+1]));
      end;
      Hf1.Init(p1, 0, count-2);
      Hf2.Init(p2, 0, count-2);
      Hf1.BestLen; // Hf1 - First byte number processing   writeln('First 8 bit.');
      Hf2.BestLen; // Hf2 - Second byte number processing  writeln('Second 8 bit.');
      Result:= hf1.Tablesize + hf2.Tablesize + (hf1.Bits + hf2.Bits + 7) div 8;

      for i:= Low(Source^) to Low(Source^)+count-2 do begin
        tempValue:= Source^[i+1]-Source^[i];
        p1^[i]:= Hi(tempValue);
        p2^[i]:= Lo(tempValue);
      end;
      Hfd1.Init(p1, 0, count-2);
      Hfd2.Init(p2, 0, count-2);
      Hfd1.BestLen; // Hf1 - First byte number processing   writeln('First 8 bit.');
      Hfd2.BestLen; // Hf2 - Second byte number processing  writeln('Second 8 bit.');
      deltas1size:= hfd1.Tablesize + hfd2.Tablesize + (hfd1.Bits + hfd2.Bits + 7) div 8;

      for i:= Low(Source^) to Low(Source^)+count-3 do begin
        tempValue:= integer(Source^[i+2]-2*Source^[i+1]+Source^[i]);
        p1^[i]:= Hi(tempValue);
        p2^[i]:= Lo(tempValue);
      end;
      Hfdd1.Init(p1, 0, count-3);
      Hfdd2.Init(p2, 0, count-3);
      Hfdd1.BestLen; // Hf1 - First byte number processing   writeln('First 8 bit.');
      Hfdd2.BestLen; // Hf2 - Second byte number processing  writeln('Second 8 bit.');
      deltas2size:= hfdd1.Tablesize + hfdd2.Tablesize + (hfdd1.Bits + hfdd2.Bits + 7) div 8;
    finally
      ReallocMem(p1, 0);
      ReallocMem(p2, 0);
    end;

    if (deltas1size<=Result) and (deltas1size<=deltas2size) then UseDeltas:= 1 else
    if (deltas2size<=Result) and (deltas2size<=deltas1size) then UseDeltas:= 2;

    case UseDeltas  of
     1: begin
          Result:= deltas1size;
          hf1.Free;
          hf2.Free;
          hfdd1.Free;
          hfdd2.Free;
          hf1:= hfd1;
          hf2:= hfd2;
        end;
     2: begin
          Result:= deltas2size;
          hf1.Free;
          hf2.Free;
          hfd1.Free;
          hfd2.Free;
          hf1:= hfdd1;
          hf2:= hfdd2;
        end;
    else begin
          hfd1.Free;
          hfd2.Free;
          hfdd1.Free;
          hfdd2.Free;
        end;
    end;

    //   
    ReallocMem(Dest, (Result+BegOfDescrTable+$100)*SizeOf(Dest^[0]));
    FillChar(Dest^, (Result+BegOfDescrTable+$100)*SizeOf(Dest^[0]), 0);

    {   }
    point:= 0;
    // 
    for i:= 1 to Length(Signature) do WriteByte(Byte(Signature[i]));
    point:= TypeOfData;
    if Compression then WriteByte(Compressed) {  }
                   else WriteByte(Stored);
    WriteCardinal(DataCount, count);
    point:= DataWidth; WriteByte(InpDataTypeSize*8);
    point:= Delta; WriteByte(UseDeltas);
    WriteWord(FirstWord, word(Source^[Low(Source^)]));

    if (Compression) then begin
      WriteCardinal(Table1Beg, BegOfDescrTable);
      point:= ReadCardinal(Table1Beg);
      inc(point, Descrtable);
      {for hf1 writing tables}
      with hf1 do begin
        {counting number of lengthes in alphabet}
        {filling desctable & decodtable}
        lencount:= 0; oldlen:= 0;
        for i:= HighIndex downto LowIndex do //  
          if oldlen<>len[i] then begin
            WriteByte(len[i]);
            inc(lencount);
            oldlen:= len[i];
          end;
        {NumOfLenghts=lencount}
        Dest^[ReadWord(Table1Beg)+NumOfLenghts]:= byte(lencount);
        {filling decode table}
        for i:= HighIndex downto LowIndex do writebyte(value[i]);
      end;
      point1:= point;
      WriteCardinal(Table2Beg, point1);
      point:= point1;
      inc(point, Descrtable);

      { for hf2 writing tables}
      with hf2 do begin
        {counting number of lengthes in alphabet}
        {filling desctable & decodtable}
        lencount:= 0; oldlen:= 0;
        for i:= HighIndex downto LowIndex do
          if oldlen<>len[i] then begin
            WriteByte(len[i]);
            inc(lencount);
            oldlen:= len[i];
          end;
        Dest^[ReadWord(Table2Beg)+NumOfLenghts]:= byte(lencount);
        {filling decode table}
        for i:= HighIndex downto LowIndex do writebyte(value[i]);
      end;
    end;
    point1:= point;
    WriteCardinal(DataBeg, point1);
    point:= point1;

    bitpointer:= 7;  pointb:= point;
   {  }
    if (Compression) then begin
      case UseDeltas of
        0:  for i:= 1 to count-1 do begin //  1      !
              with hf1 do begin
                pos:= valuePos[Hi(Source^[i])];
                WriteBits(code[pos], len[pos]);
              end;
              with hf2 do begin
                pos:= valuePos[Lo(Source^[i])];
                WriteBits(code[pos], len[pos]);
              end;
            end;
        1:  //  
            for i:= 1 to count-1 do begin //  1      !
              tempValue:= integer(Source^[i]-Source^[i-1]);
              with hf1 do begin
                pos:= valuePos[Hi(tempValue)];
                WriteBits(code[pos], len[pos]);
              end;
              with hf2 do begin
                pos:= valuePos[Lo(tempValue)];
                WriteBits(code[pos], len[pos]);
              end;
            end;
        2:  //  
            begin
              WriteBits(InpDataType(Source^[Low(Source^)+1]-Source^[Low(Source^)]), InpDataTypeSize*8);
              for i:= 1 to count-2 do begin //  1      !
                tempValue:= integer(Source^[i+1]-2*Source^[i]+Source^[i-1]);
                with hf1 do begin
                  pos:= valuePos[Hi(tempValue)];
                  WriteBits(code[pos], len[pos]);
                end;
                with hf2 do begin
                  pos:= valuePos[Lo(tempValue)];
                  WriteBits(code[pos], len[pos]);
                end;
              end;
            end;
      end;
    end
    else
      for i:= 1 to count-1 do
        WriteBits((Source^[i] shr (16-Dest^[DataWidth])), Dest^[DataWidth]);

    Result:= pointb+1;
  finally
    hf1.Free;
    hf2.Free;
  end;
end;

// *****************************************************************************
{ UnCompressor }

function  UnCompressArray(var Source: ByteArrayPtr; var Dest: SmallIntArrayPtr): InpDataLen;
var
  pointb: cardinal;
  bitpointer: shortint;

  DBeg, DCount,
  T1Beg, T2Beg: InpDataLen;

  function ReadWord(pointr: cardinal):word;
  begin
    Result:= Source^[pointr]+Source^[pointr+1]*$100;
  end;

  function ReadCardinal(pointr: cardinal):cardinal;
  begin
    Result:= Source^[pointr]+Source^[pointr+1]*$100+Source^[pointr+2]*$10000+Source^[pointr+3]*$1000000;
  end;

  function ReadNextBit(var Res: cardinal): boolean;
  begin
    Result:= false;
    Res:= Res*2;
    if (((1 shl bitpointer) and Source^[pointb]) <> 0) then begin
      Res:= Res+1;
      Result:= true;
    end;
    dec(bitpointer);
    if (bitpointer<0) then begin
      inc(pointb);
      bitpointer:= 7;
    end;
  end;

  function ReadNextBits: cardinal;  //    ,    <16 
  var i: byte;
  begin
    Result:= 0;
    for i:= 1 to Source^[DataWidth] do
      ReadNextBit(Result);
  end;

  function ReadBits(table: integer): byte;
  var
     CodeTablPointer,
     TablePointer: InpDataLen;
     TableCount: byte;
     Code: cardinal;
     DeltaBit, PrevNbit,
     Nbit: byte;
     All_1s: boolean;
  begin
    Result:= 0;
    if table=1 then TablePointer:= T1Beg
               else TablePointer:= T2Beg;
    Code:= 0; PrevNbit := 0;
    CodeTablPointer:= TablePointer + DescrTable + Source^[TablePointer]; {  }
    All_1s:= true;
    for TableCount:= 1 to Source^[TablePointer] do begin {    }
      inc(CodeTablPointer, Code);
      Code:= 0;
      DeltaBit:= Source^[TablePointer+TableCount]-PrevNbit;
      for Nbit:= 1 to DeltaBit do
        All_1s:= ReadNextBit(Code) and All_1s;{  }
      if (not All_1s) or (TableCount=Source^[TablePointer]) then begin
        {    1     ,     }
        Result:= Source^[CodeTablPointer + Code];
        break;
      end
      else begin
        {     N }
        PrevNbit:= Source^[TablePointer+TableCount];
      end;
    end;
  end;

var
  i: integer;
  UseDeltas: byte;
begin
  Result:= 0;
  if not Assigned(Source) then exit;

  { }
  DBeg:= ReadCardinal(DataBeg);
  DCount:= ReadCardinal(DataCount);
  T1Beg:= ReadCardinal(Table1Beg);
  T2Beg:= ReadCardinal(Table2Beg);
  pointb:= DBeg; bitpointer:= 7;
  ReallocMem(Dest, DCount*sizeof(Dest^[0]));
  Result:= DCount;
  if Result<1 then exit;

  UseDeltas:= Source^[Delta];

  Dest^[0]:= ReadWord(FirstWord);
  if (Source^[TypeOfData]=Compressed) then
    case UseDeltas of
      0:  for i:= 1 to DCount-1 do begin
            Dest^[i]:= InpDataType(word(ReadBits(1)*$100+ReadBits(2)))
          end;
      1:  for i:= 1 to DCount-1 do begin
            Dest^[i]:= Dest^[i-1] + InpDataType(word(ReadBits(1)*$100+ReadBits(2)))
          end;
      2:  begin
            Dest^[1]:= Dest^[0] + InpDataType(word(ReadNextBits));
            for i:= 2 to DCount-1 do begin
              Dest^[i]:= 2*Dest^[i-1] - Dest^[i-2]+ InpDataType(word(ReadBits(1)*$100+ReadBits(2)))
            end;
          end;
    end
  else
    for i:= 1 to DCount-1 do
      Dest^[i]:= InpDataType(ReadNextBits shl (16 - Source^[DataWidth]));

end;

end.
